


"""TODO.

"""



# validate_edge_widths(tree, 0)
# validate_edge_widths(tree, 1)
# validate_edge_widths(tree, 1.555)
# validate_edge_widths(tree, np.arange(tree.nnodes))
# validate_edge_widths(tree, ("dist", ))
# validate_edge_widths(tree, ("dist", 2))
# validate_edge_widths(tree, ("dist", 2, 4))
# validate_edge_widths(tree, ("dist", 2, 4, 0))