#!/usr/bin/env python

"""
Organization and API access of the pcm subpackage.

The modules in :mod:`toytree.pcm.src` contain Classes that are
exposed in user-facing factory functions written and documented
in :mod:`toytree.pcm.src.api`. The API docs are built from `api`,

Examples
--------
>>> toytree.pcm.get_vcv()
>>> toytree.pcm.get_qmatrix()

>>> toytree.pcm.sim_discrete_markov_data()
>>> toytree.pcm.sim_continuous_data()

>>> toytree.pcm.div_equal_splits()
>>> toytree.pcm.div_tip_level()
>>> toytree.pcm.div_tip_level()
"""

# from .discrete_markov_model_fit import simulate_discrete_markov_data
