#!/usr/bin/env python

"""Educational resources submodule.

IDEAS
-----
- tree traversal animations
- wright_fisher: simulation, animation, and pair w/ tree visualizations
- markov: [todo] markov model visualization w/ animations.
- substitution: Markov model sympy demonstration tools.
- ...

"""

from toytree.learn.src.wrightfisher import wright_fisher_simulation
